import operator
import typing

from listful import MoreThanOneResultException, NotFoundException
from listful.exceptions import ListfulsMismatchException
from listful.index import Index, SimpleIndex
from listful.types import FIELD, GETTER, ITEM, VALUE
from listful.utils import intersect_lists, naive_find_elements_in_list


class Listful(typing.List[ITEM], typing.Generic[ITEM, VALUE]):
    def __init__(
        self,
        iterable: typing.Iterable[ITEM],
        fields: typing.List[FIELD],
        getter: typing.Optional[GETTER[ITEM, VALUE]] = None,
    ):
        super().__init__(iterable)
        self.fields = fields
        if getter is None:
            # pylint: disable=unsubscriptable-object
            if len(self) > 0 and isinstance(self[0], dict):
                getter = typing.cast(GETTER[ITEM, VALUE], operator.getitem)
            else:
                getter = getattr
        self.getter: GETTER[ITEM, VALUE] = getter
        self._indexes: typing.Dict[FIELD, Index[ITEM, VALUE]] = {}
        self._build_indexes()

    def _build_indexes(self) -> None:
        self._indexes = {}
        for field in self.fields:
            index = self._indexes[field] = SimpleIndex()
            for element in self:  # pylint: disable=not-an-iterable
                value = self.getter(element, field)
                index.add(element, value)

    def filter(self, **kwargs: typing.Any) -> 'Results[ITEM, VALUE]':
        results = None
        for field, value in kwargs.items():
            if field in self._indexes:
                new_results = self._indexes[field].get(value)
            else:
                new_results = naive_find_elements_in_list(
                    self, field, value, self.getter
                )
            if results is None:
                results = new_results
            else:
                results = intersect_lists(results, new_results)
        if results is None:
            results = self
        return Results(results, filter_=kwargs, source=self)

    def get_all_for_field(self, field: FIELD) -> typing.List[VALUE]:
        return [
            self.getter(element, field)
            for element in self  # pylint: disable=not-an-iterable
        ]

    def rebuild_indexes_for_item(self, item: ITEM) -> None:
        for field in self.fields:
            value = self.getter(item, field)
            self._indexes[field].add(item, value)

    def append(self, item: ITEM) -> None:
        super().append(item)  # pylint: disable=no-member
        self.rebuild_indexes_for_item(item)

    def _remove_item_from_indexes(self, item: ITEM) -> None:
        for field in self.fields:
            value = self.getter(item, field)
            self._indexes[field].remove(item, value)

    def remove(self, item: ITEM) -> None:
        super().remove(item)  # pylint: disable=no-member
        self._remove_item_from_indexes(item)

    @typing.overload
    def __setitem__(self, index: int, item: ITEM) -> None:
        ...  # pragma: nocover

    @typing.overload
    def __setitem__(self, index: slice, item: typing.Iterable[ITEM]) -> None:
        ...  # pragma: nocover

    def __setitem__(
        self,
        index: typing.Union[int, slice],
        item: typing.Union[ITEM, typing.Iterable[ITEM]],
    ) -> None:
        self._remove_indexes_from_multiple_items(index)
        if isinstance(index, slice):
            items = typing.cast(typing.Iterable[ITEM], item)
            super().__setitem__(  # pylint: disable=no-member
                typing.cast(slice, index), items
            )
            for item_ in items:
                self.rebuild_indexes_for_item(item_)
        else:
            item = typing.cast(ITEM, item)
            super().__setitem__(  # pylint: disable=no-member
                typing.cast(int, index), item
            )
            self.rebuild_indexes_for_item(item)

    def __delitem__(self, index: typing.Union[int, slice]) -> None:
        self._remove_indexes_from_multiple_items(index)
        super().__delitem__(index)  # pylint: disable=no-member

    def _remove_indexes_from_multiple_items(
        self, index: typing.Union[int, slice]
    ) -> None:
        item = self[index]  # pylint: disable=unsubscriptable-object
        if isinstance(index, slice):
            items = typing.cast(typing.Iterable[ITEM], item)
            for item_ in items:
                self._remove_item_from_indexes(item_)
        else:
            item = typing.cast(ITEM, item)
            self._remove_item_from_indexes(item)

    def extend(self, iterable: typing.Iterable[ITEM]) -> None:
        for element in iterable:
            self.append(element)

    @classmethod
    def from_listfuls(
        cls, listfuls: typing.Iterable['Listful[ITEM, VALUE]']
    ) -> 'Listful[ITEM, VALUE]':
        listfuls_iterator = iter(listfuls)
        try:
            result = next(listfuls_iterator)
        except StopIteration:
            raise ValueError('Expected at least one listful object')
        fields = set(result.fields)
        getter = result.getter
        for listful_obj in listfuls_iterator:
            if set(listful_obj.fields) != fields:
                raise ListfulsMismatchException(
                    f"Can't merge listfuls with different indexes {listful_obj.fields} != {fields}"
                )
            if listful_obj.getter != getter:
                raise ListfulsMismatchException(
                    f"Can't merge listfuls with different getters {listful_obj.getter} != {getter}"
                )
            result.extend(listful_obj)
        return result


class Results(typing.Generic[ITEM, VALUE]):
    def __init__(
        self,
        results: typing.Sequence[ITEM],
        filter_: typing.Dict[FIELD, VALUE],
        source: 'Listful[ITEM, VALUE]',
    ):
        self._source = source
        self._filter = filter_
        self._results = results

    def one_or_none(self) -> typing.Optional[ITEM]:
        if len(self._results) == 1:
            return next(iter(self._results))
        return None

    def one_or_raise(self) -> ITEM:
        results = self._results
        if len(results) == 1:
            return next(iter(results))
        if len(results) > 1:
            raise MoreThanOneResultException(
                f'Found more than one result for filter {self._filter}: {results}'
            )
        raise NotFoundException(
            f'Found zero results for filter {self._filter}'
        )

    def to_list(self) -> typing.List[ITEM]:
        return list(self._results)

    def to_listful(self) -> 'Listful[ITEM, VALUE]':
        return Listful(self._results, self._source.fields)



import typing


class Item:
    def __init__(self, x: int, y: int):
        self.x = x
        self.y = y

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, Item):
            return self.x == other.x and self.y == other.y
        return False



import typing

T = typing.TypeVar('T')
V = typing.TypeVar('V')
F = typing.TypeVar('F')


def intersect_lists(
    list1: typing.Iterable[T], list2: typing.Iterable[T]
) -> typing.Iterable[T]:
    return [element for element in list1 if element in list2]


def naive_find_elements_in_list(
    elements: typing.List[T],
    field: F,
    value: V,
    getter: typing.Callable[[T, F], V],
) -> typing.List[T]:
    return [element for element in elements if getter(element, field) == value]



import collections
import typing

from listful.types import ITEM, VALUE


class Index(typing.Generic[ITEM, VALUE]):
    # pylint: disable=unused-argument
    def add(self, element: ITEM, value: VALUE) -> None:
        ...  # pragma: nocover

    # pylint: disable=unused-argument
    def remove(self, element: ITEM, value: VALUE) -> None:
        ...  # pragma: nocover

    # pylint: disable=unused-argument
    def get(self, value: VALUE) -> typing.List[ITEM]:
        ...  # pragma: nocover


class SimpleIndex(Index[ITEM, VALUE]):
    # pylint: disable=super-init-not-called
    def __init__(self) -> None:
        self._index: typing.Dict[
            VALUE, typing.List[ITEM]
        ] = collections.defaultdict(list)

    def add(self, element: ITEM, value: VALUE) -> None:
        self._index[value].append(element)

    def remove(self, element: ITEM, value: VALUE) -> None:
        self._index[value].remove(element)

    def get(self, value: VALUE) -> typing.List[ITEM]:
        return self._index[value]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['listful', 'listful._internal']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'listful',
    'version': '0.1.3',
    'description': 'efficient filtering of lists of objects',
    'long_description': '# listful\n\n[![pypi](https://badge.fury.io/py/listful.svg)](https://pypi.org/project/listful)\n[![Python: 3.6+](https://img.shields.io/badge/Python-3.6+-blue.svg)](https://pypi.org/project/listful)\n[![Downloads](https://img.shields.io/pypi/dm/listful.svg)](https://pypistats.org/packages/listful)\n[![Build Status](https://travis-ci.org/d1618033/listful.svg?branch=master)](https://travis-ci.org/d1618033/listful)\n[![Code coverage](https://codecov.io/gh/d1618033/listful/branch/master/graph/badge.svg)](https://codecov.io/gh/d1618033/listful)\n[![License: MIT](https://img.shields.io/badge/License-MIT-green.svg)](https://en.wikipedia.org/wiki/MIT_License)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n## Description\n\nEfficient filtering of lists of objects\n\n## Installation\n\n    pip install listful\n\n## Usage\n\n\nInitialize with the fields you want to filter by:\n```\n>>> from listful import Listful\n>>> data = Listful(\n...    [{\'x\': 1, \'y\': 10}, {\'x\': 2, \'y\': 20}, {\'x\': 2, \'y\': 30}], \n...    fields=[\'x\', \'y\']\n... )\n```\n\n### Filtering:\n\n* By one field:\n```\n>>> data.filter(x=1).one_or_none()\n{\'x\': 1, \'y\': 10}\n>>> data.filter(y=20).one_or_none()\n{\'x\': 2, \'y\': 20}\n```\n\n* By one field, with more than one result:\n```\n>>> data.filter(x=2).to_list()\n[{\'x\': 2, \'y\': 20}, {\'x\': 2, \'y\': 30}]\n```\n\n* By two fields:\n```\n>>> data.filter(x=2, y=30).one_or_none()\n{\'x\': 2, \'y\': 30}\n```\n\n* Raise exception if more than one found\n``` \n>>> data.filter(x=2).one_or_raise()\nTraceback (most recent call last):\n<...>\nlistful.exceptions.MoreThanOneResultException: Found more than one result for filter {\'x\': 2}: [{\'x\': 2, \'y\': 20}, {\'x\': 2, \'y\': 30}]\n```\n\n* Get all values for a specific field\n\n```\n>>> data.get_all_for_field(\'x\')\n[1, 2, 2]\n```\n\n### Updating indexes:\n\n`Listful` has the same api as `list`, so you can get/set/delete items the same way \nand the indices will be updated automatically\n\n```\n>>> data[0] = {\'x\': 17, \'y\': 17}\n>>> data.filter(x=17).one_or_none()\n{\'x\': 17, \'y\': 17}\n>>> data[0]\n{\'x\': 17, \'y\': 17}\n>>> del data[0]\n>>> data.filter(x=17).one_or_none()\n``` \n\nIf you want to modify an element and update the indices you can do so explicitly:\n```\n>>> data[0][\'x\'] = 1\n>>> data.rebuild_indexes_for_item(data[0])\n>>> data.filter(x=1).one_or_none()\n{\'x\': 1, \'y\': 20}\n``` \n\n\n### Objects:\n\nListful supports also lists of objects:\n\n```\n>>> class Item:\n...     def __init__(self, x, y):\n...         self.x = x\n...         self.y = y\n...\n...     def __repr__(self):\n...         return f"Item(x={self.x}, y={self.y})"\n\n>>> items = Listful(\n...    [Item(x=1, y=10), Item(x=2, y=20), Item(x=2, y=30)], \n...    fields=[\'x\', \'y\']\n... )\n>>> items.filter(x=1).one_or_none()\nItem(x=1, y=10)\n```\n## For developers\n\n### Create venv and install deps\n\n    make init\n\n### Install git precommit hook\n\n    make precommit_install\n\n### Run linters, autoformat, tests etc.\n\n    make pretty lint test\n\n### Bump new version\n\n    make bump_major\n    make bump_minor\n    make bump_patch\n\n## License\n\nMIT\n\n## Change Log\n\nUnreleased\n-----\n\n* ...\n\n0.1.3 - 2020-02-14\n-----\n\n* ...\n\n0.1.1 - 2020-02-12\n-----\n\n* ...\n\n0.1.0 - 2020-02-12\n-----\n\n* initial\n',
    'author': 'David S',
    'author_email': 'd1618033@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/listful',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
